/*****************************************************************************/
/*                                                                           */
/*   LCD Back Light Control with AVR                                         */
/*                                                                           */
/*   (c) 2008 Dipl.-Ing. (FH) Rainer Reusch, Reusch Elektronik for Elektor   */
/*   http://www.reusch-elektronik.de                                         */
/*   http://reweb.fh-weingarten.de/elektor                                   */
/*                                                                           */
/*   Atmel AVR Studio V4.13.528                                              */
/*   GNU C-Compiler GCC V4.2.20                                              */
/*                                                                           */
/*   Created: 2008-02-19                                                     */
/*   Changed: 2008-02-21                                                     */
/*                                                                           */
/*****************************************************************************/

#include <avr/interrupt.h>
#include "backlight.h"

// internal used variables
uint16_t  Measure;				// summary of measured current value (for average calculation)
uint8_t  	MeasureCount;		// counter

//-----------------------------------------------------------------------------
void SetPWM(uint8_t pwm)
//-----------------------------------------------------------------------------
// set the PWM (inverted, low active)
{
	PWMValue=pwm;
	if (pwm)
	{	// PWM on
		TCCR2=(1<<WGM21)|(1<<WGM20)|(1<<COM21)|(1<<COM20)|(1<<CS20);	// fast PWM (WGM), inverted PWM (COM), no clock prescaling
		OCR2=pwm;
	}
	else
	{	// PWM off
		TCCR2=0;
		PWMPort|=(1<<PWMPin);  // pin to high level
	}
}

//-----------------------------------------------------------------------------
ISR(ADC_vect)
//-----------------------------------------------------------------------------
{	// ISR: AD conversion complete
  // regulate current
	Measure+=ADCH;	// get measured value (only 8 Bit)
	MeasureCount=(MeasureCount+1)&(NumAvrgVals-1);
	if (!MeasureCount)
	{
		// calculate average value
		Measure=(Measure+(NumAvrgVals>>1))>>NumAvrgPwr;
		// regulate current
		if (Measure>Current)
		// current regulation (integral regulation)
		{	// current to high -> decrement PWM
			if (PWMValue) SetPWM(PWMValue-1);
		}
		else
		if (Measure<Current)
		{	// current to low -> increment PWM
			if (PWMValue<0xFF) SetPWM(PWMValue+1);
		}
		Measure=0;
	}
}

//-----------------------------------------------------------------------------
void Init_Backlight(uint8_t TurnOnPWM)
//-----------------------------------------------------------------------------
// initialize backlight control without feedback
// call this routine only one time at startup
{
	// init PWM port
	PWMPort|=1<<PWMPin;	// set PWM output to high level (backlight current off)
	PWMDir|=1<<PWMPin;	// set PWM port as output
	// init timer and PWM
	ASSR&=~(1<<AS2);	// timer clock source to I/O clock
	TCCR2=0;	// PWM off
	SetPWM(TurnOnPWM);
}

//-----------------------------------------------------------------------------
void Init_BacklightFB(uint8_t TurnOnCurrent)
//-----------------------------------------------------------------------------
// initialize backlight control with feedback
// call this routine only one time at startup
{
	Current=TurnOnCurrent;
	PWMValue=0;
	Measure=0;
	MeasureCount=0;
	// init analogue feedback input
	FBDir&=~(1<<FBPin);	// set as input
	// init PWM port
	PWMPort|=1<<PWMPin;	// set PWM output to high level (backlight current off)
	PWMDir|=1<<PWMPin;	// set PWM port as output
	// init timer and PWM
	ASSR&=~(1<<AS2);	// timer clock source to I/O clock
	TCCR2=0;	// PWM off
	// A/D converter
	ADMUX=(1<<REFS1)|(1<<REFS0)|(1<<ADLAR);  // internal reference, left adjust result, single ended input at ADC0
	ADCSRA=(1<<ADEN)|(1<<ADSC)|(1<<ADATE)|(1<<ADIE)|(1<<ADPS2)|(1<<ADPS1)|(1<<ADPS0);  // enable ADC, start conversion, enable free running, enable interrupt, prescaler 128
	SFIOR&=~((1<<ADTS2)|(1<<ADTS1)|(1<<ADTS0));	// ADC in free running mode
}

